"""
pip install ecdsa
"""

from hashlib import sha256
from ecdsa.numbertheory import jacobi, square_root_mod_prime
from ecdsa import NIST256p
from ecdsa.ellipticcurve import Point
import os

curve = NIST256p.curve

# --------------- helper functions ----------------

def get_good_random(size):
    return int.from_bytes(os.urandom(size), 'big')

# NIST approved DRBG algorithm
# https://projectbullrun.org/dual-ec/documents/SP800-90A.pdf#page=68
class DualEC():

    def __init__(self):
        self.P = NIST256p.generator
        self.Q = 0x74686973206973207468652072696768742070617468 * self.P
        self.state = get_good_random(32)
    
    def get_random(self):
        r = (self.state * self.P).x()
        self.state =  (r * self.P).x()
        out = (r * self.Q).x()

        # removing the first 16 bits is overrated
        return out

# gotta spice sha256 up
def spicy_sha256(input):
    mask = bin(int.from_bytes(input, 'big'))[2:]
    mask = mask.zfill((len(mask) + 7) // 8 * 8)
    out = 0

    for i in range(len(mask)):
        preimage = f"hsbp_camp2024_{i}"

        # reverse the preimage
        if mask[i] == "1":
            preimage = preimage[::-1]
        
        # hint: xor is isomorphic to ???
        out ^= int.from_bytes(sha256(preimage.encode()).digest(), 'big')
    
    return out

# classic try-and-increment method
# https://eprint.iacr.org/2009/226.pdf#page=2
def hash_to_curve(msg):
    # Who needs the full hash?
    x = (spicy_sha256(msg) << 32) & (2**256 - 1)
    a, b, p = curve.a(), curve.b(), curve.p()

    while True:
        y2 = (x**3 + a*x + b) % p
        
        if jacobi(y2, p) == 1:
            break

        x += 1
    
    y = square_root_mod_prime(y2, p)
    H = Point(curve, x, y)

    # Custom modification to make the result appear more random. 
    return 0x656c6c697074696320637572766520616c6765627261 * H

# ------------- OPAQUE implementation -------------

# OPRF blinding, calculates H(pw) ^ r
# https://datatracker.ietf.org/doc/html/draft-irtf-cfrg-opaque#section-5.2.1
def create_registration_request(password, csprng):
    H = hash_to_curve(password)
    r = csprng.get_random()

    return r * H

# the server knows how to cook good random numbers
def create_registration_response(blinded):
    server_kU = get_good_random(32) # per-user private key
    server_sk = get_good_random(32) # long-term private key

    evaluated = server_kU * blinded
    server_pk = server_sk * NIST256p.generator

    return server_kU, server_sk, evaluated, server_pk

# ------------------- challenge -------------------

FLAG = os.getenv("FLAG", "camp2024{this_is_a_fake_flag_12345}")

assert FLAG.startswith("camp2024{") and FLAG.endswith("}")

csprng = DualEC()
password = FLAG[9:-1].encode()

assert len(password) == 25

# start opaque
# client-side
uid = csprng.get_random()
blinded = create_registration_request(password, csprng)

print(f"[CreateRegistrationRequest] C -> S: uid = {uid}, blinded = {str(blinded)}")

# server-side
server_kU, server_sk, evaluated, server_pk = create_registration_response(blinded)

print(f"[CreateRegistrationResponse] S -> C: evaluated = {str(evaluated)}, server_pk = {str(server_pk.to_affine())}")

# TODO: implement the other parts of the OPAQUE algorithm
